"""
BrainRender Wrapper
===================

:author: Yu-Ting Wei

This module provides a CLI-based wrapper for `brainrender <https://brainglobe.info/documentation/brainrender/index.html>`_.
Once installed, the CLI can be invoked directly from the command line.

To see available options, use the ``-h`` flag:

.. prompt:: bash $

    neuralib_brainrender {area, roi, probe} -h

----

Region Reconstruction
---------------------

Example: Reconstructing the Visual Cortex from specified brain regions:

.. prompt:: bash $

    neuralib_brainrender area -R VISal,VISam,VISl,VISli,VISp,VISpl,VISpm,VISpor --camera top

|brender area|

----

ROI Reconstruction
------------------

By default, coordinates are interpreted in the CCF coordinate space.
You can specify the coordinate space using the ``--coord-space`` option: ``{ccf, brainrender}``.

**NumPy File Input**
^^^^^^^^^^^^^^^^^^^^

Input shape: ``Array[float, (N, 3)]``, with AP, DV, and ML coordinates.

Example:

.. code-block:: python

    [[-3.03,  4.34, -4.50],
     [-3.03,  4.42, -4.37],
     ...
     [-2.91,  4.12,  4.85]]

Run:

.. prompt:: bash $

    neuralib_brainrender roi --file <NUMPY_FILE>

**CSV File Input**
^^^^^^^^^^^^^^^^^^

Required columns: ``AP_location``, ``DV_location``, ``ML_location``

.. code-block:: text

    ┌─────────────┬─────────────┬─────────────┐
    │ AP_location │ DV_location │ ML_location │
    │------------ │-------------│-------------│
    │ -3.03       │ 4.34        │ -4.50       │
    │ -3.03       │ 4.92        │ -4.31       │
    │ ...         │ ...         │ ...         │
    │ -2.91       │ 4.12        │ 4.85        │
    └─────────────┴─────────────┴─────────────┘


Example:

.. prompt:: bash $

    neuralib_brainrender roi --file <CSV_FILE>


|brender roi|



**Flexible Reconstruction (Processed CSV)**
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Be able to reconstruct rois in a specific regions/subregions


Example of using parsed allenccf csv output ::

    ┌───────────────────────────────────┬─────────┬─────────────┬─────────────┬─────────────┬─────────┬─────────┬────────┬────────────┐
    │ name                              ┆ acronym ┆ AP_location ┆ DV_location ┆ ML_location ┆ avIndex ┆ channel ┆ source ┆  ...       │
    │ ---                               ┆ ---     ┆ ---         ┆ ---         ┆ ---         ┆ ---     ┆ ---     ┆ ---    ┆  ---       │
    │ str                               ┆ str     ┆ f64         ┆ f64         ┆ f64         ┆ i64     ┆ str     ┆ str    ┆  ...       │
    ╞═══════════════════════════════════╪═════════╪═════════════╪═════════════╪═════════════╪═════════╪═════════╪════════╪════════════╡
    │ Ectorhinal area/Layer 5           ┆ ECT5    ┆ -3.03       ┆ 4.34        ┆ -4.5        ┆ 377     ┆ gfp     ┆ VIS    ┆  ...       │
    │ Perirhinal area layer 6a          ┆ PERI6a  ┆ -3.03       ┆ 4.42        ┆ -4.37       ┆ 372     ┆ gfp     ┆ VIS    ┆  ...       │
    │ …                                 ┆ …       ┆ …           ┆ …           ┆ …           ┆ …       ┆ …       ┆ …      ┆  …         │
    │ Ventral auditory area layer 6a    ┆ AUDv6a  ┆ -2.91       ┆ 3.52        ┆ 4.46        ┆ 156     ┆ rfp     ┆ CA1    ┆  ...       │
    │ Ectorhinal area/Layer 6a          ┆ ECT6a   ┆ -2.91       ┆ 4.14        ┆ 4.47        ┆ 378     ┆ rfp     ┆ CA1    ┆  ...       │
    │ Temporal association areas layer… ┆ TEa5    ┆ -2.91       ┆ 4.02        ┆ 4.55        ┆ 365     ┆ rfp     ┆ CA1    ┆  ...       │
    └───────────────────────────────────┴─────────┴─────────────┴─────────────┴─────────────┴─────────┴─────────┴────────┴────────────┘

.. code-block:: python

    import polars as pl
    from neuralib.atlas.ccf.classifier import RoiClassifierDataFrame

    df = pl.DataFrame({
        "acronym": ["RSPd", "RSPd", "VISp", "VISp"],
        "AP_location": [1.2, 1.3, -2.4, -2.6],
        "DV_location": [1.0, 1.1, 2.0, 2.1],
        "ML_location": [0.4, -0.3, 0.2, -0.2],
        "channel": ["gfp", "gfp", "rfp", "rfp"],
        "source": ["CA1", "CA1", "CA3", "CA3"]
    })

    df = RoiClassifierDataFrame(df).post_processing().dataframe()
    df.write_csv(CSV_FILE)


.. seealso::

    :class:`~neuralib.atlas.ccf.dataframe.RoiClassifierDataFrame`


Example (reconstruct ROI in the parahippocampal areas):

.. prompt:: bash $

    neuralib_brainrender roi --classifier-file <CSV_FILE> --region APr,ENT,HATA,PAR,POST,PRE,ProS,SUB --roi-region RHP --region-alpha 0.2 --roi-radius 20 --no-root -H right


|brender roi_region|



----

Probe Reconstruction
--------------------

Reconstruct probes (or shanks) based on trajectory labeling (e.g., DiI, DiO, or lesion tracks).

- Default coordinate space: CCF
- Set coordinate space using: ``--coord-space {ccf, brainrender}``
- Each shank must have 2 points: dorsal and ventral

**NumPy File Input**
^^^^^^^^^^^^^^^^^^^^

Single shank: ``Array[float, (2, 3)]`` (dorsal and ventral 3D AP/ML/DV coordinates)

.. code-block:: python

    [[-3.82, 1.92, -3.12],
     [-3.93, 4.36, -3.30]]

Multi-shank: ``Array[float, (S, 2, 3)]``

.. code-block:: python

    [[[...], [...]],
     [[...], [...]],
     ...]

**CSV File Input**
^^^^^^^^^^^^^^^^^^

Required fields: ``AP_location``, ``DV_location``, ``ML_location``, ``point``, ``probe_idx``

If loss either ``point``, ``probe_idx`` field, then auto infer based on the given insertion ``--plane``

.. code-block:: text

    ┌─────────────┬─────────────┬─────────────┬─────────┬───────────┐
    │ AP_location │ DV_location │ ML_location │ point   │ probe_idx │
    ├─────────────┼─────────────┼─────────────┼─────────┼───────────┤
    │ -3.81       │ 1.92        │ -3.12       │ dorsal  │ 1         │
    │ -3.93       │ 4.36        │ -3.30       │ ventral │ 1         │
    │ ...         │ ...         │ ...         │ ...     │ ...       │
    └─────────────┴─────────────┴─────────────┴─────────┴───────────┘

**Additional Options**
^^^^^^^^^^^^^^^^^^^^^^

- ``--depth DEPTH``: Depth (in µm) of the implantation from the brain surface
- ``--dye``: Only reconstruct dye-labeled tracks (default includes both dye and theoretical)
- ``--remove-outside-brain``: Exclude any segments outside the brain

Example: Reconstructing a 4-shank NeuroPixel probe targeting the left entorhinal cortex

.. prompt:: bash $

    neuralib_brainrender probe -F <FILE> --depth 3000 -P sagittal -R ENT -H left

- Red = dye-labeled track
- Black = theoretical track

|brender probe|

----

Help
----

To explore available options for each subcommand, use the ``-h`` flag:

.. prompt:: bash $

    neuralib_brainrender probe -h



.. |brender area| image:: ../_static/brender_area.png
.. |brender roi| image:: ../_static/brender_roi.png
.. |brender roi_region| image:: ../_static/brender_roi_regions.png
.. |brender probe| image:: ../_static/brender_probe.png


"""



from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
from iblatlas.atlas import AllenAtlas
from iblatlas.regions import BrainRegions
from matplotlib.axes import Axes

from neuralib.io.core import IBL_CACHE_DIRECTORY
from neuralib.typing import ArrayLike, PathLike, ArrayLikeStr
from neuralib.util.deprecation import deprecated_class

__all__ = [
    'IBLAtlasPlotWrapper'
]

IBL_PLANE_TYPE = Literal['coronal', 'sagittal', 'horizontal', 'top']
IBL_MAPPING_TYPE = Literal['Allen', 'Beryl', 'Cosmos', 'Swanson']  # merge areas
HEMISPHERE_TYPE = Literal['left', 'right', 'both']
IBL_BG_TYPE = Literal['image', 'boundary']


@deprecated_class(new='basic case already covered natively, otherwise, use iblatlas directly', removal_version='0.5.0')
class IBLAtlasPlotWrapper(AllenAtlas):
    """Wrapper for iblatlas plotting usage

    .. seealso:: `<https://int-brain-lab.github.io/iblenv/atlas_examples.html>`_
    """
    regions: BrainRegions

    def __init__(self, res_um: int = 10,
                 scaling: tuple[int, int, int] = (1, 1, 1),
                 mock: bool = False,
                 source_path: PathLike | None = None,
                 alpha: float = 1.0):
        """

        :param res_um: The Atlas resolution in micrometres; one of 10, 25 or 50um
        :param scaling: Scale factor along ml, ap, dv for squeeze and stretch (default: [1, 1, 1])
        :param mock: For testing purposes, return atlas object with image comprising zeros
        :param source_path: The location of the image volume. By default, use ``IBL_CACHE_DIRECTORY``
        :param alpha:
        """

        if source_path is None:
            source_path = IBL_CACHE_DIRECTORY

        super().__init__(res_um, scaling, mock, source_path)

        self._alpha = alpha

    def get_acronym_list(self, mapping: IBL_MAPPING_TYPE) -> list[str]:
        """get acronym list from different mapping type

        :param mapping: `IBL_MAPPING_TYPE`
        """
        idx = np.unique(self.regions.mappings[mapping])
        # noinspection PyUnresolvedReferences
        return self.regions.acronym[idx]
        # return list(self.regions.acronym[self.regions.mappings[mapping]])

    def plot_scalar_on_slice(
            self,
            regions: ArrayLikeStr, *,
            values: ArrayLike | None = None,
            coord: int = -570,
            plane: IBL_PLANE_TYPE = 'coronal',
            mapping: IBL_MAPPING_TYPE = 'Allen',
            hemisphere: HEMISPHERE_TYPE = 'left',
            background: IBL_BG_TYPE = 'image',
            cmap: str = 'viridis',
            clevels: ArrayLike | None = None,
            show_cbar: bool = False,
            empty_color: str = 'silver',
            ax: Axes | None = None,
            vector: bool = False,
            slice_files: np.ndarray | None = None,
            auto_merge: bool = True,
            verbose: bool = True,
            **kwargs
    ) -> tuple[plt.Figure, Axes] | tuple[plt.Figure, Axes, plt.colorbar]:
        """
        Plot the slice view.

        See detail in ``iblatlas.plots.plot_scalar_on_slice``

        .. seealso:: `<https://int-brain-lab.github.io/iblenv/notebooks_external/atlas_plotting_scalar_on_slice.html>`_

        :param regions: ArrayLike[str, R]
        :param values: ArrayLike[float, R]
        :param coord:
        :param plane: ``IBL_PLANE_TYPE``. {'coronal', 'sagittal', 'horizontal', 'top'}
        :param mapping: ``IBL_MAPPING_TYPE``. {'Allen', 'Beryl', 'Cosmos', 'Swanson'}
        :param hemisphere: ``HEMISPHERE_TYPE``. {'left', 'right', 'both'}
        :param background: ``IBL_BG_TYPE``. {'image', 'boundary'}
        :param cmap:
        :param clevels:
        :param show_cbar:
        :param empty_color:
        :param ax:
        :param vector:
        :param slice_files:
        :param auto_merge: merge for the certain mapping
        :param verbose:
        :param kwargs:
        :return:
        """

        from iblatlas.plots import plot_scalar_on_slice

        if not isinstance(regions, np.ndarray):
            regions = np.array(regions)

        if values is None:
            values = np.arange(regions.size)

        if auto_merge:
            regions = np.unique(self.regions.acronym2acronym(regions, mapping=mapping))

            if any([r not in self.get_acronym_list(mapping) for r in regions]):
                raise RuntimeError('auto merging error')

            values = np.arange(regions.size)

        if verbose:
            print(f'Plot region: {list(regions)}')

        # noinspection PyTypeChecker
        return plot_scalar_on_slice(
            regions, values, coord,
            slice=plane,
            mapping=mapping,
            hemisphere=hemisphere,
            background=background,
            cmap=cmap,
            clevels=clevels,
            show_cbar=show_cbar,
            empty_color=empty_color,
            brain_atlas=self,
            ax=ax,
            vector=vector,
            slice_files=slice_files,
            **kwargs
        )

    def plot_points_on_slice(
            self,
            xyz: np.ndarray,
            values: ArrayLike | None = None,
            coord: int = -570,
            plane: IBL_PLANE_TYPE = 'coronal',
            mapping: IBL_MAPPING_TYPE = 'Allen',
            background: IBL_BG_TYPE = 'boundary',
            cmap: str = 'Reds',
            clevels: ArrayLike | None = None,
            show_cbar: bool = True,
            aggr: Literal['sum', 'count', 'mean', 'std', 'median', 'min', 'max'] = 'count',
            fwhm: int = 0,
            ax: Axes | None = None,
    ) -> tuple[plt.Figure, Axes] | tuple[plt.Figure, Axes, plt.colorbar]:
        """
        plot_points_on_slice

        :param xyz:
        :param values:
        :param coord:
        :param plane:
        :param mapping:
        :param background:
        :param cmap:
        :param clevels:
        :param show_cbar:
        :param aggr:
        :param fwhm:
        :param ax:
        :return:
        """
        from iblatlas.plots import plot_points_on_slice

        return plot_points_on_slice(
            xyz, values, coord, plane, mapping,
            background, cmap, clevels, show_cbar, aggr, fwhm,
            self, ax
        )

    def _plot_slice(self, im, extent, ax=None, cmap=None, volume=None, **kwargs):
        """Overwrite by pass other kwargs"""
        if ax is None:
            ax = plt.gca()
            ax.axis('equal')

        if cmap is None:
            cmap = plt.get_cmap('bone')

        if volume == 'boundary':
            imb = np.zeros((*im.shape[:2], 4), dtype=np.uint8)
            imb[im == 1] = np.array([0, 0, 0, 255])
            im = imb

        ax.imshow(im, extent=extent, cmap=cmap, alpha=self._alpha)



"""
WideField
=========

:author:
    Yu-Ting Wei

This module provides tools for processing widefield video sequence data and is currently under active development.

Two interactive tools are currently available:



FFT View
--------

A Bokeh-based visualization of the HSV colormap representation of the Fourier-transformed video (e.g., for visual retinotopy).

Usage:

.. code-block:: python

    neuralib_widefield fft <FILE>

|fft_view|



Align
------

A Napari-based tool for aligning video data to a dorsal cortex view.

Usage:

.. code-block:: python

    neuralib_widefield align <FILE> [-R REFERENCE_FILE] [-M MAP_FILE]

|align_view|



.. |fft_view| image:: ../_static/example_fft.jpg
.. |align_view| image:: ../_static/example_align.jpg


"""
from .fft import *
from .plot import *
from .svd import *



from typing import TypedDict

import polars as pl

__all__ = ['plot_task_gantt']


class GanttDict(TypedDict, total=False):
    task: str
    start: str
    finish: str
    resource: str


def plot_task_gantt(jobs: list[GanttDict]):
    """
    plot gantt
    * Example

    >>> x = [GanttDict(task='task1', start='2023-09-01', finish='2023-11-30'),
    ...      GanttDict(task='task2', start='2024-01-01', finish='2024-09-30')]
    >>> plot_task_gantt(x)

    :param jobs: list of ``GanttDict``
    """
    import plotly.express as px

    df = pl.DataFrame(jobs)
    if 'resource' not in df.columns:
        color = 'task'
    else:
        color = 'resource'
    fig = px.timeline(df, x_start='start', x_end='finish', y='task', color=color)
    fig.update_layout(
        width=1200,
        height=400
    )
    fig.show()



from __future__ import annotations

from pathlib import Path
from typing import NamedTuple, Any

import attrs
import numpy as np
import polars as pl
from scipy.io import loadmat
from scipy.io.matlab import MatlabOpaque

from neuralib.atlas.typing import PLANE_TYPE
from neuralib.atlas.view import SlicePlane, get_slice_view
from neuralib.typing import PathLike

__all__ = [
    'CCFTransMatrix',
    'load_transform_matrix'
]


def load_transform_matrix(filepath: PathLike,
                          plane_type: PLANE_TYPE,
                          resolution: int = 10,
                          default_name: str = 'test') -> CCFTransMatrix:
    """
    matrix for image transformation

    :param filepath: transformation matrix .mat file
    :param plane_type: ``PLANE_TYPE`` {'coronal', 'sagittal', 'transverse'}
    :param resolution: atlas resolution
    :param default_name: name of the ``CCFTransMatrix``
    :return: ``CCFTransMatrix``
    """
    mat = loadmat(filepath, squeeze_me=True, struct_as_record=False)['save_transform']
    f = Path(filepath).name

    try:
        name = f[:f.index('_resize')]
    except ValueError:
        name = default_name

    return CCFTransMatrix(
        name,
        MatMatrix(mat.allen_location, mat.transform, mat.transform_points),
        plane_type,
        resolution=resolution
    )


class MatMatrix(NamedTuple):
    allen_location: np.ndarray
    """abnormal parsing in loadmat. [996 array([  0, -12], dtype=int16)]"""
    transform: MatlabOpaque
    """mat meta info. currently not used"""
    transform_points: np.ndarray
    """[(P, 2), (P, 2)] transformation point for 2dccf GUI site view"""

    @property
    def slice_index(self) -> int:
        return int(self.allen_location[0])

    # noinspection PyTypeChecker
    @property
    def delta_values(self) -> tuple[int, int]:
        """shifting values for yx axis (height, width) for a specific plane type.

        Coronal slice: [0]:DV, [1]:ML

        Sagittal Slice: [0]:DV, [1]:AP

        """
        delta = self.allen_location[1]
        return delta[0], delta[1]


@attrs.frozen
class CCFTransMatrix:
    """matrix for image transformation"""

    slice_id: str
    """slice name id"""
    matrix: MatMatrix
    """MatMatrix"""
    plane_type: PLANE_TYPE
    """PLANE_TYPE"""
    resolution: int = attrs.field(kw_only=True, default=10)
    """resolution in um"""

    @property
    def slice_index(self) -> int:
        """slice index"""
        return self.matrix.slice_index

    @property
    def delta_xy(self) -> tuple[int, int]:
        """map delta value to xy slice view"""
        return self.matrix.delta_values[1], self.matrix.delta_values[0]

    def get_slice_plane(self) -> SlicePlane:
        """get slice plane"""
        ret = (
            get_slice_view('reference', self.plane_type, resolution=self.resolution)
            .plane_at(self.slice_index)
        )

        dw = self.delta_xy[0]
        dh = self.delta_xy[1]

        ret = ret.with_offset(
            dw=dw + 1 if dw != 0 else 0,  # avoid index err
            dh=dh + 1 if dh != 0 else 0,  # avoid index err
        )

        return ret

    def matrix_info(self, to_polars: bool = True) -> dict[str, Any] | pl.DataFrame:
        """
        transform matrix information dict

        :param to_polars: to polars dataframe
        :return:
        """
        ret = {
            'slice_id': self.slice_id,
            'plane_type': self.plane_type,
            'atlas_resolution': self.resolution,
            'slice_index': self.slice_index,
            'reference_value': self.get_slice_plane().reference_value,
            'dw': self.delta_xy[0],
            'dh': self.delta_xy[1]
        }

        if to_polars:
            ret = pl.DataFrame(ret)

        return ret



"""
Validator Usage Guide
=====================

Overview
--------
This guide demonstrates how to use the validator builders provided by our library. Each section
provides short examples followed by a reference table of the builder’s methods.

.. contents::
   :local:
   :depth: 2


String Validation
=================

Examples
--------
**Minimum String Length**::

    from neuralib.argp import validator

    class Opt:
        # Must be at least 2 characters long
        a: str = argument('-a', validator.str.length_in_range(2, None))

    opt = Opt()
    opt.a = 'Hi'    # OK
    opt.a = ''      # Raises ValueError

**Regex Matching**::

    class Opt:
        # Must match a letter followed by a digit, e.g. 'a1', 'b9'
        a: str = argument('-a', validator.str.match(r'[a-z][0-9]'))

    opt = Opt()
    opt.a = 'a1'    # OK
    opt.a = 'A1'    # Raises ValueError

Method Reference
----------------
.. list-table::
   :header-rows: 1
   :widths: 30 70

   * - **Method**
     - **Description**
   * - :meth:`length_in_range(a, b) <StrValidatorBuilder.length_in_range>`
     - Enforces a string length in [a, b]. Either bound may be ``None``.
   * - :meth:`match(pattern) <StrValidatorBuilder.match>`
     - Checks if the string matches a given regex pattern.
   * - :meth:`starts_with(prefix) <StrValidatorBuilder.starts_with>`
     - Checks if the string starts with ``prefix``.
   * - :meth:`ends_with(suffix) <StrValidatorBuilder.ends_with>`
     - Checks if the string ends with ``suffix``.
   * - :meth:`contains(substring) <StrValidatorBuilder.contains>`
     - Checks if the string contains the given substring.
   * - :meth:`is_in(options) <StrValidatorBuilder.is_in>`
     - Checks if the string is in the provided collection of allowed options.


Integer Validation
==================

Examples
--------
**Integer Range**::

    class Opt:
        # Must be >= 2
        a: int = argument('-a', validator.int.in_range(2, None))

    opt = Opt()
    opt.a = 5   # OK
    opt.a = 0   # Raises ValueError

**Positivity**::

    class Opt:
        # Must be strictly positive
        a: int = argument('-a', validator.int.positive(include_zero=False))

    opt = Opt()
    opt.a = 10  # OK
    opt.a = 0   # Raises ValueError

Method Reference
----------------
.. list-table::
   :header-rows: 1
   :widths: 30 70

   * - **Method**
     - **Description**
   * - :meth:`in_range(a, b) <IntValidatorBuilder.in_range>`
     - Checks if integer is in [a, b]. Either bound may be ``None``.
   * - :meth:`positive(include_zero=True) <IntValidatorBuilder.positive>`
     - Checks if integer is >= 0 (if ``include_zero=True``) or > 0 otherwise.
   * - :meth:`negative(include_zero=True) <IntValidatorBuilder.negative>`
     - Checks if integer is <= 0 (if ``include_zero=True``) or < 0 otherwise.


Float Validation
================

Examples
--------
**Range + NaN Handling**::

    class Opt:
        # Must be < 100, NaN not allowed
        a: float = argument('-a',
            validator.float.in_range(None, 100).allow_nan(False)
        )

    opt = Opt()
    opt.a = 3.14        # OK
    opt.a = 123.45      # Raises ValueError (out of range)
    opt.a = float('nan')# Raises ValueError (NaN not allowed)

Method Reference
----------------
.. list-table::
   :header-rows: 1
   :widths: 30 70

   * - **Method**
     - **Description**
   * - :meth:`in_range(a, b) <FloatValidatorBuilder.in_range>`
     - Checks if float is in the open interval ``(a, b)``.
   * - :meth:`in_range_closed(a, b) <FloatValidatorBuilder.in_range_closed>`
     - Checks if float is in the closed interval ``[a, b]``.
   * - :meth:`allow_nan(allow=True) <FloatValidatorBuilder.allow_nan>`
     - Allows or disallows NaN values.
   * - :meth:`positive(include_zero=True) <FloatValidatorBuilder.positive>`
     - Checks if float is >= 0 (if ``include_zero=True``) or > 0 otherwise.
   * - :meth:`negative(include_zero=True) <FloatValidatorBuilder.negative>`
     - Checks if float is <= 0 (if ``include_zero=True``) or < 0 otherwise.


List Validation
===============

Examples
--------
**List of Integers**::

    class Opt:
        # Must be a list of integers
        a: list[int] = argument('-a', validator.list(int))

    opt = Opt()
    opt.a = [1, 2, 3]    # OK
    opt.a = ['a', 2]     # Raises ValueError

**Item Validation**::

    class Opt:
        # Each item must be non-negative
        a: list[int] = argument('-a',
            validator.list(int).on_item(validator.int.positive(True))
        )

    opt = Opt()
    opt.a = [0, 2, 5]    # OK
    opt.a = [1, -1]      # Raises ValueError

Method Reference
----------------
.. list-table::
   :header-rows: 1
   :widths: 30 70

   * - **Method**
     - **Description**
   * - :meth:`length_in_range(a, b) <ListValidatorBuilder.length_in_range>`
     - Enforces list length in [a, b].
   * - :meth:`allow_empty(allow=True) <ListValidatorBuilder.allow_empty>`
     - Allows or disallows an empty list.
   * - :meth:`on_item(validator) <ListValidatorBuilder.on_item>`
     - Applies a validator to each list item.


Tuple Validation
================

Examples
--------
**Fixed-Length Tuple**::

    class Opt:
        # Must be (str, int, float)
        a: tuple[str, int, float] = argument(
            '-a', validator.tuple(str, int, float)
        )

    opt = Opt()
    opt.a = ('abc', 42, 3.14)   # OK
    opt.a = ('abc', 42)        # Raises ValueError (too few elements)

**Variable-Length**::

    class Opt:
        # Must be (str, int, ...) i.e. at least 'str + int', optionally more ints
        a: tuple[str, int, ...] = argument(
            '-a', validator.tuple(str, int, ...)
        )

    opt = Opt()
    opt.a = ('x', 10)            # OK
    opt.a = ('x', 10, 20, 30)    # OK
    opt.a = ('x',)               # Raises ValueError (missing int)

**Item-Validation**::

    class Opt:
        # Must be (str, int, float).
        # The string must have a length <= 5,
        # and the int must be >= 0 and <= 100.
        a: tuple[str, int, float] = argument(
            '-a',
            validator.tuple(str, int, float)
                .on_item(0, validator.str.length_in_range(None, 5))
                .on_item(1, validator.int.in_range(0, 100))
        )

    opt = Opt()

    # Passes all checks: str length=3, int in range [0..100], float is fine
    opt.a = ('hey', 42, 3.14)

    # Fails because the string is too long:
    opt.a = ('excessive', 42, 1.2)
    # Raises ValueError: str length over 5: "excessive"

    # Fails because integer is out of range:
    opt.a = ('hi', 999, 2.5)
    # Raises ValueError: value out of range [0, 100]: 999

Method Reference
----------------
.. list-table::
   :header-rows: 1
   :widths: 30 70

   * - **Method**
     - **Description**
   * - :meth:`on_item(indexes, validator) <TupleValidatorBuilder.on_item>`
     - Apply a validator to specific tuple positions, or ``None`` for all.
   * - *(constructor)*
     - Pass one int (e.g. 3) to enforce a fixed-length tuple with no type checks, or a tuple of types
       like ``(str, int, float)``. The last type can be ``...`` for variable length.


Path Validation
=================

Examples
--------
**Path suffix**::

    class Opt:
        a: Path = argument('-a', validator.path.is_suffix(['.csv', '.npy']))

    opt = Opt()
    opt.a = Path('.../*.csv')    # OK
    opt.a = Path('.../*.txt')    # Raises ValueError


Method Reference
----------------
.. list-table::
   :header-rows: 1
   :widths: 30 70

   * - **Method**
     - **Description**
   * - :meth:`is_suffix(suffix) <PathValidatorBuilder.is_suffix>`
     - Check path suffix or in a list of suffixes
   * - :meth:`is_exists() <PathValidatorBuilder.is_exists>`
     - Check if path exists
   * - :meth:`is_file() <PathValidatorBuilder.is_file>`
     - Check if path is a file
   * - :meth:`is_dir() <PathValidatorBuilder.is_dir>`
     - Check if path is a directory




Logical Combinators
===================

Examples
--------
**OR Combination**::

    class Opt:
        # Must be int in [0,10] OR str length in [0,10]
        a: int | str = argument(
            '-a',
            validator.any(
                validator.int.in_range(0, 10),
                validator.str.length_in_range(0, 10)
            )
        )

    opt = Opt()
    opt.a = 5            # OK (int in [0..10])
    opt.a = 'abc'        # OK (length=3)
    opt.a = 50           # Raises ValueError

**AND Combination**::

    class Opt:
        # Must be non-negative AND non-positive => zero
        a: int = argument('-a', validator.all(
            validator.int.positive(include_zero=True),
            validator.int.negative(include_zero=True)
        ))

    opt = Opt()
    opt.a = 0   # OK
    opt.a = 1   # Raises ValueError
    opt.a = -1  # Raises ValueError

Method Reference
----------------
.. list-table::
   :header-rows: 1
   :widths: 30 70

   * - **Method/Class**
     - **Description**
   * - :meth:`validator.any(...) <ValidatorBuilder.any>` or ``|``
     - Combine validators with logical OR; passing at least one is enough.
   * - :meth:`validator.all(...) <ValidatorBuilder.all>` or ``&``
     - Combine validators with logical AND; must pass them all.
   * - ``OrValidatorBuilder``
     - The class implementing OR logic.
   * - ``AndValidatorBuilder``
     - The class implementing AND logic.


Error Handling
--------------
If any validation fails:
- A :class:`ValidatorFailError` (or subclass) is raised, often rethrown as ``ValueError``
in higher-level frameworks.

- **Type mismatches** specifically raise :class:`ValidatorFailOnTypeError`.


"""
from __future__ import annotations

import re
from collections.abc import Callable
from pathlib import Path
from typing import Any, TypeVar, Generic, final, overload, Collection

from typing_extensions import Self

T = TypeVar('T')


class ValidatorFailError(ValueError):
    pass


class ValidatorFailOnTypeError(ValidatorFailError):
    """
    A special ValidatorFailError that is raised when type validation failure.
    It is used for Validator#any() to exclude some error message.
    """
    pass


class Validator:
    def __call__(self, value: Any) -> bool:
        return True


class LambdaValidator(Validator, Generic[T]):
    def __init__(self, validator: Callable[[T], bool],
                 message: str | Callable[[T], str] = None):
        if isinstance(message, str):
            message = message.__mod__

        self.__validator = validator
        self.__message = message

    def __call__(self, value: T) -> bool:
        message = self.__message
        try:
            success = self.__validator(value)
        except ValidatorFailError:
            raise
        except BaseException as e:
            if message is None:
                raise ValidatorFailError('validate failure') from e
            else:
                raise ValidatorFailError(message(value)) from e
        else:
            if success is None or success:
                return True
            elif message is None:
                return False
            else:
                raise ValidatorFailError(message(value))

    def __and__(self, validator: Callable[[Any], bool]) -> AndValidatorBuilder:
        return AndValidatorBuilder(self, validator)

    def __or__(self, validator: Callable[[Any], bool]) -> OrValidatorBuilder:
        return OrValidatorBuilder(self, validator)


@final
class ValidatorBuilder:
    @property
    def str(self) -> StrValidatorBuilder:
        return StrValidatorBuilder()

    @property
    def int(self) -> IntValidatorBuilder:
        return IntValidatorBuilder()

    @property
    def float(self) -> FloatValidatorBuilder:
        return FloatValidatorBuilder()

    @overload
    def tuple(self, element_type: int) -> TupleValidatorBuilder:
        pass

    @overload
    def tuple(self, *element_type: type[T]) -> TupleValidatorBuilder:
        pass

    # noinspection PyMethodMayBeStatic
    def tuple(self, *element_type) -> TupleValidatorBuilder:
        return TupleValidatorBuilder(element_type)

    # noinspection PyMethodMayBeStatic
    def list(self, element_type: type[T] = None) -> ListValidatorBuilder:
        return ListValidatorBuilder(element_type)

    @property
    def path(self):
        return PathValidatorBuilder()

    @classmethod
    def all(cls, *validator: Callable[[T], bool]) -> AndValidatorBuilder:
        return AndValidatorBuilder(*validator)

    @classmethod
    def any(cls, *validator: Callable[[T], bool]) -> OrValidatorBuilder:
        return OrValidatorBuilder(*validator)

    @classmethod
    def optional(cls) -> Validator:
        return LambdaValidator(lambda it: it is None)

    @classmethod
    def non_none(cls) -> Validator:
        return LambdaValidator(lambda it: it is not None)

    def __call__(self, validator: Callable[[Any], bool],
                 message: str | Callable[[Any], str] = None) -> LambdaValidator:
        return LambdaValidator(validator, message)


class AbstractTypeValidatorBuilder(Validator, Generic[T]):
    def __init__(self, value_type: type[T] | tuple[type[T], ...] = None):
        self.__value_type = value_type
        self.__validators: list[LambdaValidator[T]] = []
        self.__allow_none = False

    def __call__(self, value: Any) -> bool:
        if value is None:
            if self.__allow_none:
                return True
            else:
                raise ValidatorFailError('None')

        # noinspection PyTypeHints
        if self.__value_type is not None and not isinstance(value, self.__value_type):
            raise ValidatorFailOnTypeError(f'not instance of {self.__value_type.__name__} : {value}')

        for validator in self.__validators:
            if not validator(value):
                return False

        return True

    @overload
    def _add(self, validator: LambdaValidator[T]):
        pass

    @overload
    def _add(self, validator: Callable[[T], bool], message: str | Callable[[T], str] = None):
        pass

    def _add(self, validator, message=None):
        if not isinstance(validator, LambdaValidator):
            validator = LambdaValidator(validator, message)
        self.__validators.append(validator)

    def optional(self) -> Self:
        self.__allow_none = True
        return self

    def __and__(self, validator: Callable[[Any], bool]) -> AndValidatorBuilder:
        return AndValidatorBuilder(self, validator)

    def __or__(self, validator: Callable[[Any], bool]) -> OrValidatorBuilder:
        return OrValidatorBuilder(self, validator)


class StrValidatorBuilder(AbstractTypeValidatorBuilder[str]):
    def __init__(self):
        super().__init__(str)

    def length_in_range(self, a: int | None, b: int | None, /) -> StrValidatorBuilder:
        """Enforce a string length range"""
        match (a, b):
            case (int(a), None):
                self._add(lambda it: a <= len(it), f'str length less than {a}: "%s"')
            case (None, int(b)):
                self._add(lambda it: len(it) <= b, f'str length over {b}: "%s"')
            case (int(a), int(b)):
                self._add(lambda it: a <= len(it) <= b, f'str length out of range [{a}, {b}]: "%s"')
            case _:
                raise TypeError()
        return self

    def match(self, r: str | re.Pattern) -> StrValidatorBuilder:
        """Check if string matches a regular expression"""
        if isinstance(r, str):
            r = re.compile(r)
        self._add(lambda it: r.match(it) is not None, f'str does not match to {r.pattern} : "%s"')
        return self

    def starts_with(self, prefix: str) -> StrValidatorBuilder:
        """Check if string values start with a substring"""
        self._add(lambda it: it.startswith(prefix), f'str does not start with {prefix}: "%s"')
        return self

    def ends_with(self, suffix: str) -> StrValidatorBuilder:
        """Check if string values end with a substring"""
        self._add(lambda it: it.endswith(suffix), f'str does not end with {suffix}: "%s"')
        return self

    def contains(self, text: str) -> StrValidatorBuilder:
        """Check if string values contain a substring"""
        self._add(lambda it: text in it, f'str does not contain {text}: "%s"')
        return self

    def is_in(self, options: Collection[str]) -> StrValidatorBuilder:
        """Check if string is one of the allow options"""
        self._add(lambda it: it in options, f'str not in allowed set {options}: "%s"')
        return self


class IntValidatorBuilder(AbstractTypeValidatorBuilder[int]):
    def __init__(self):
        super().__init__(int)

    def in_range(self, a: int | None, b: int | None, /) -> IntValidatorBuilder:
        """Enforce a numeric range for int values"""
        match (a, b):
            case (int(a), None):
                self._add(lambda it: a <= it, f'value less than {a}: %d')
            case (None, int(b)):
                self._add(lambda it: it <= b, f'value over {b}: %d')
            case (int(a), int(b)):
                self._add(lambda it: a <= it <= b, f'value out of range [{a}, {b}]: %d')
            case _:
                raise TypeError()

        return self

    def positive(self, include_zero=True):
        """Check if an int value is positive or non-negative"""
        if include_zero:
            self._add(lambda it: it >= 0, 'not a non-negative value : %d')
        else:
            self._add(lambda it: it > 0, 'not a positive value : %d')
        return self

    def negative(self, include_zero=True):
        """Check if an int value is negative or non-positive."""
        if include_zero:
            self._add(lambda it: it <= 0, 'not a non-positive value : %d')
        else:
            self._add(lambda it: it < 0, 'not a negative value : %d')
        return self


class FloatValidatorBuilder(AbstractTypeValidatorBuilder[float]):
    def __init__(self):
        super().__init__((int, float))
        self.__allow_nan = False

    def in_range(self, a: float | None, b: float | None, /) -> Self:
        """Enforce an open-interval numeric range (a < value < b)"""
        match (a, b):
            case (int(a), None):
                self._add(lambda it: a < it, f'value less than {a}: %f')
            case (None, int(b)):
                self._add(lambda it: it < b, f'value over {b}: %f')
            case (int(a), int(b)):
                self._add(lambda it: a < it < b, f'value out of range ({a}, {b}): %f')
            case _:
                raise TypeError()

        return self

    def in_range_closed(self, a: float | None, b: float | None, /) -> Self:
        """ Enforce a closed-interval numeric range (a <= value <= b)"""
        match (a, b):
            case (int(a), None):
                self._add(lambda it: a <= it, f'value less than {a}: %f')
            case (None, int(b)):
                self._add(lambda it: it <= b, f'value over {b}: %f')
            case (int(a), int(b)):
                self._add(lambda it: a <= it <= b, f'value out of range [{a}, {b}]: %f')
            case _:
                raise TypeError()
        return self

    def allow_nan(self, allow: bool = True) -> Self:
        """Allow or disallow NaN (not a number) as a valid float"""
        self.__allow_nan = allow
        return self

    def positive(self, include_zero=True) -> Self:
        """Check if a float value is positive or non-negative"""
        if include_zero:
            self._add(lambda it: it >= 0, 'not a non-negative value: %f')
        else:
            self._add(lambda it: it > 0, 'not a positive value: %f')
        return self

    def negative(self, include_zero=True) -> Self:
        """Check if a float value is negative or non-positive"""
        if include_zero:
            self._add(lambda it: it <= 0, 'not a non-positive value : %f')
        else:
            self._add(lambda it: it < 0, 'not a negative value : %f')
        return self

    def __call__(self, value: Any) -> bool:
        if value != value:
            if self.__allow_nan:
                return True
            else:
                raise ValidatorFailError('NaN')

        return super().__call__(value)


class ListValidatorBuilder(AbstractTypeValidatorBuilder[list[T]]):
    def __init__(self, element_type: type[T] = None):
        super().__init__()
        self.__element_type = element_type
        self.__allow_empty = True

    def length_in_range(self, a: int | None, b: int | None, /) -> Self:
        """Enforce a length range for lists"""
        match (a, b):
            case (int(a), None):
                self._add(lambda it: a <= len(it),
                          lambda it: f'list length less than {a}: {len(it)}')
            case (None, int(b)):
                self._add(lambda it: len(it) <= b,
                          lambda it: f'list length over {b}: {len(it)}')
            case (int(a), int(b)):
                self._add(lambda it: a <= len(it) <= b,
                          lambda it: f'list length out of range [{a}, {b}]: {len(it)}')
            case _:
                raise TypeError()

        return self

    def allow_empty(self, allow: bool = True):
        """Allow or disallow empty lists"""
        self.__allow_empty = allow

    def on_item(self, validator: Callable[[Any], bool]) -> Self:
        """Apply an additional validator to each item in the list

        :param validator: A callable that validates each item
        """
        self._add(ListItemValidatorBuilder(validator))
        return self

    def __call__(self, value: Any) -> bool:
        if not isinstance(value, (tuple, list)):
            raise ValidatorFailOnTypeError(f'not a list : {value}')

        if not self.__allow_empty and len(value) == 0:
            raise ValidatorFailError(f'empty list : {value}')

        if (element_type := self.__element_type) is not None:
            for i, element in enumerate(value):
                if not element_isinstance(element, element_type):
                    raise ValidatorFailError(f'wrong element type at {i} : {element}')

        return super().__call__(value)


class TupleValidatorBuilder(AbstractTypeValidatorBuilder[tuple]):
    def __init__(self, element_type: tuple[int] | tuple[type[T], ...]):
        super().__init__()

        match element_type:
            case (int(length), ):
                element_type = (None,) * length

        self.__element_type = element_type

    def on_item(self, item: int | list[int] | None, validator: Callable[[Any], bool]) -> Self:
        """Apply a validator to specific tuple positions

        :param item: A single index, a list of indices, or None for all indices
        :param validator: The validation callable to apply
        """
        if item is None:
            pass
        elif isinstance(item, int):
            if item < 0:
                raise ValueError('should always use positive index')
        else:
            for index in item:
                if index < 0:
                    raise ValueError('should always use positive index')

        # check range
        if item is not None and len(self.__element_type) > 0:
            if isinstance(item, int):
                et = self.__element_type[item]
                if et is ...:
                    raise IndexError()
            else:
                for index in item:
                    et = self.__element_type[index]
                    if et is ...:
                        raise IndexError()

        self._add(TupleItemValidatorBuilder(item, validator))

        return self

    def __call__(self, value: Any) -> bool:
        if not isinstance(value, tuple):
            raise ValidatorFailOnTypeError(f'not a tuple : {value}')

        if len(element_type := self.__element_type) > 0:
            if element_type[-1] is ...:
                at_least_length = len(element_type) - 1
                if len(value) < at_least_length:
                    raise ValidatorFailError(f'length less than {at_least_length} : {value}')

                for i, e, t in zip(range(at_least_length), value, element_type):
                    if t is not None and not element_isinstance(e, t):
                        raise ValidatorFailError(f'wrong element type at {i} : {e}')

                if at_least_length > 0:
                    last_element_type = element_type[at_least_length - 1]
                    for i, e in zip(range(at_least_length, len(value)), value[at_least_length:]):
                        if not element_isinstance(e, last_element_type):
                            raise ValidatorFailError(f'wrong element type at {i} : {e}')

            else:
                if len(value) != len(element_type):
                    raise ValidatorFailError(f'length not match to {len(element_type)} : {value}')

                for i, e, t in zip(range(len(element_type)), value, element_type):
                    if t is not None and not element_isinstance(e, t):
                        raise ValidatorFailError(f'wrong element type at {i} : {e}')

        return super().__call__(value)


class PathValidatorBuilder(AbstractTypeValidatorBuilder[Path]):

    def __init__(self):
        super().__init__(Path)

    def is_suffix(self, suffix: str | list[str] | tuple[str, ...]) -> Self:
        """Check path suffix or in a list of suffixes"""
        if isinstance(suffix, str):
            self._add(lambda it: it.suffix == suffix, f'suffix != {suffix}: %s')
        elif isinstance(suffix, list | tuple):
            self._add(lambda it: it.suffix in suffix, f'suffix not in {suffix}: %s')
        else:
            raise TypeError('')

        return self

    def is_exists(self) -> Self:
        """Check if path exists"""
        self._add(lambda it: it.exists(), f'path does not exist: %s')
        return self

    def is_file(self) -> Self:
        """Check if path is a file"""
        self._add(lambda it: it.is_file(), f'path is not a file: %s')
        return self

    def is_dir(self) -> Self:
        """Check if path is a directory"""
        self._add(lambda it: it.is_dir(), f'path is not a directory: %s')
        return self


class ListItemValidatorBuilder(LambdaValidator):
    def __call__(self, value: Any) -> bool:
        for i, element in enumerate(value):
            try:
                fail = not super().__call__(element)
            except BaseException as e:
                raise ValidatorFailError(f'at index {i}, ' + e.args[0]) from e
            else:
                if fail:
                    raise ValidatorFailError(f'at index {i}, validate fail : {value}')
        return True


class TupleItemValidatorBuilder(LambdaValidator):
    def __init__(self, item: int | list[int] | None, validator: Callable[[Any], bool]):
        super().__init__(validator)
        self.__item = item

    def __call__(self, value: Any) -> bool:
        if self.__item is None:
            for index in range(len(value)):
                if not self.__call_on_index__(index, value):
                    return False
            return True
        elif isinstance(self.__item, int):
            return self.__call_on_index__(self.__item, value)
        else:
            for index in self.__item:
                if not self.__call_on_index__(index, value):
                    return False
            return True

    def __call_on_index__(self, index: int, value: Any) -> bool:
        try:
            element = value[index]
        except IndexError as e:
            raise ValidatorFailError(f'index {index} out of size {len(value)}') from e

        try:
            return super().__call__(element)
        except BaseException as e:
            raise ValidatorFailError(f'at index {index}, ' + e.args[0]) from e


class OrValidatorBuilder(Validator):
    def __init__(self, *validator: Callable[[Any], bool]):
        self.__validators = list(validator)

    def __call__(self, value: Any) -> bool:
        if len(self.__validators) == 0:
            return True

        coll = []
        for validator in self.__validators:
            try:
                if validator(value):
                    return True
            except ValidatorFailOnTypeError:
                pass
            except BaseException as e:
                if len(e.args):
                    coll.append(e.args[0])

        raise ValidatorFailError('; '.join(coll))

    def __and__(self, validator: Callable[[Any], bool]) -> AndValidatorBuilder:
        return AndValidatorBuilder(self, validator)

    def __or__(self, validator: Callable[[Any], bool]) -> OrValidatorBuilder:
        if isinstance(validator, OrValidatorBuilder):
            self.__validators.extend(validator.__validators)
        else:
            self.__validators.append(validator)
        return self


class AndValidatorBuilder(Validator):
    def __init__(self, *validator: Callable[[Any], bool]):
        self.__validators = list(validator)

    def __call__(self, value: Any) -> bool:
        if len(self.__validators) == 0:
            return True

        for validator in self.__validators:
            if not validator(value):
                raise ValidatorFailError()

        return True

    def __and__(self, validator: Callable[[Any], bool]) -> AndValidatorBuilder:
        if isinstance(validator, AndValidatorBuilder):
            self.__validators.extend(validator.__validators)
        else:
            self.__validators.append(validator)
        return self

    def __or__(self, validator: Callable[[Any], bool]) -> OrValidatorBuilder:
        return OrValidatorBuilder(self, validator)


def element_isinstance(e, t) -> bool:
    if isinstance(t, type):
        return isinstance(e, t)

    if t is Any:
        return True

    if callable(t):
        try:
            return True if t(e) else False
        except TypeError:
            return False

    print(f'NotImplementedError(element_isinstance(..., {t}))')
    return False



"""
Annotation-based argparse
=========================

:author:
    Ta-Shun Su

This module provide a way to integrate python ``argparse`` module into class attribute
and annotation type, that allow options have type information, and allow parser combination
easily.

class Argument and function argument
------------------------------------

A simple option class that contains several options carried by their attributes.

>>> class ExampleOptions:
...     ANIMAL: str = argument('--ANIMAL')
...     EXP_DATE: str = argument('--EXP_DATE')
...     OUTPUT_DIR: str = argument()

In ``ExampleOptions``, ``ANIMAL`` is an attribute with type annotation ``str``. it has
a class variable :class:`~neuralib.argp.core.Argument` (:func:`~neuralib.argp.core.argument()` return) which contains the
arguments of ``ArgumentParser.add_argument``. For now, we have an optional
argument ``--ANIMAL`` which accept one argument. ``EXP_DATE`` as is another optional argument.
And ``OUTPUT_DIR`` are a positonal argument because it doesn't have dashed options.

>>> opt = parse_args(ExampleOptions())
... print(opt.ANIMAL)

After class declared, you can use :meth:`~neuralib.argp.core.parse_args()` to parse cli arguments. This
function will create an ``ArgumentParser`` and find out all argument attributes.
Then set the attributes from parsed result.

Commandline usage
~~~~~~~~~~~~~~~~~

In bash, you can call this option class ::

    python -m module.path --ANIMAL name --EXP_DATE date output

:func:`~neuralib.argp.core.print_help()` does the similar things but print the help document to the stdout.

>>> print_help(opt)

Or use ``-h`` options ::

    python -m module.path -h


Annotation type infering
~~~~~~~~~~~~~~~~~~~~~~~~

In general, you can think :func:`~neuralib.argp.core.argument` just a delegate funcion that passes the arguments
to the ``ArgumentParser.add_argument``. However, this function will try to
infer missing arguments based on the annotation type when creating the ``ArgumentParser``.
For now, this module is not powerful to handle all possible case. there are support type
(Please see :meth:`~neuralib.argp.core.Argument.complete_kwargs()` for detailed):

1. ``bool``: infer parameter ``action`` to ``store_true``, ``default`` to ``False``.
2. ``Literal[...]`` : infer parameter ``choices``.
3. ``Optional[T]``: infer parameter ``type`` to ``T`` if it is callable. If it is ``Literal``, apply 2.
4. ``callable(...)`` with signature ``Callable[[str], T]``: infer parameter ``type`` to ``T``
5. parameter ``dest`` always use the attribute name.

Additionally, :class:`~neuralib.argp.core.Argument` also provide a parameter ``group`` to reduce the complexity of
create subgrouping parser.

Option class compose
~~~~~~~~~~~~~~~~~~~~

Option class can be composed by inherition. Child option class can also change the value from parent's
argument. As well as disable it (by replacing a value)

>>> class MoreOptions(ExampleOptions):
...     # additional optional option
...     verbose: bool = argument('-v', '--verbose')
...     # change default value
...     ANIMAL: str = as_argument(ExampleOptions.animal).with_options(default='YW00')
...     # and disable an option
...     OUTPUT_DIR: str = 'output' # just replace with a value

Change options name is more complicate, because you might want to add more name, remove some name,
or rename some name. :meth:`~neuralib.argp.core.Argument.with_options()` allow you to do that:

>>> class ChangeExample(ExampleOptions):
...     # replace option name: --animal
...     ANIMAL: str = as_argument(ExampleOptions.animal).with_options('--animal')
...     # add more option name: -A, --ANIMAL
...     ANIMAL: str = as_argument(ExampleOptions.animal).with_options(..., '-A')
...     # rename option: --animal
...     ANIMAL: str = as_argument(ExampleOptions.animal).with_options({
...         '--ANIMAL': '--animal'
...     })

Utility function for option class
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

``ExampleOptions`` doesn't declated neither ``__str__`` nor ``__repr__``, so it is not convenient to debug.
funciont :func:`~neuralib.argp.core.as_dict()` provide a way to take argument attribute's value into a dictionary.

>>> as_dict(opt)
{'ANIMAL': ..., 'EXP_DATE': ..., 'OUTPUT_DIR': ...}

Option class is not restricted into only one use case. It works like a normal class.

class AbstractParser
--------------------

This class provide a main like class that has more control on ``argparse.ArgumentParser``
creation.

>>> class ExampleParser(AbstractParser, ExampleOptions):
...     DESCRIPTION = 'Example parser'
...     def run(self):
...         ...
>>> if __name__ == '__main__':
...     ExampleParser().main()

Subcommands
-----------

This module isn't fully support sub-command feature, but only provide a simple way for specific case:

>>> parse_command_args(
...     description='top level parser',
...     parsers=dict(ep=ExampleParser)
... )

In bash::

    python -m module.path ep --ANIMAL name --EXP_DATE date output


Validator usage
-----------------------
The validator system provides a fluent, chainable API for building type-specific validation rules.
It works by defining specialized “builder” classes (e.g., for strings, integers, floats, lists, and tuples)
that let you specify constraints like numeric ranges, string length ranges, regex checks, or container length/item rules

See detailed in :attr:`~neuralib.argp.validator`


"""
from ._type import *
from ._validator import *
from .core import *
